"""NumpyArrayItem.

This module defines the NumpyArrayItem class, which represents a NumPy array item.
"""

from __future__ import annotations

from functools import cached_property
from json import dumps, loads
from typing import TYPE_CHECKING

from skore.item.item import Item, ItemTypeError

if TYPE_CHECKING:
    import numpy


class NumpyArrayItem(Item):
    """
    A class to represent a NumPy array item.

    This class encapsulates a NumPy array along with its creation and update timestamps.
    """

    def __init__(
        self,
        array_json: str,
        created_at: str | None = None,
        updated_at: str | None = None,
    ):
        """
        Initialize a NumpyArrayItem.

        Parameters
        ----------
        array_json : str
            The JSON representation of the array.
        created_at : str
            The creation timestamp in ISO format.
        updated_at : str
            The last update timestamp in ISO format.
        """
        super().__init__(created_at, updated_at)

        self.array_json = array_json

    @cached_property
    def array(self) -> numpy.ndarray:
        """
        The numpy array from the persistence.

        Its content can differ from the original array because it has been serialized
        using `json.dumps` function and not pickled, in order to be
        environment-independent.
        """
        import numpy

        return numpy.asarray(loads(self.array_json))

    @classmethod
    def factory(cls, array: numpy.ndarray) -> NumpyArrayItem:
        """
        Create a new NumpyArrayItem instance from a NumPy array.

        Parameters
        ----------
        array : numpy.ndarray
            The NumPy array to store.

        Returns
        -------
        NumpyArrayItem
            A new NumpyArrayItem instance.
        """
        import numpy

        if not isinstance(array, numpy.ndarray):
            raise ItemTypeError(f"Type '{array.__class__}' is not supported.")

        return cls(array_json=dumps(array.tolist()))



"""In-memory storage."""

from collections.abc import Iterator
from typing import Any

from .abstract_storage import AbstractStorage


class InMemoryStorage(AbstractStorage):
    """In-memory storage."""

    def __init__(self):
        """
        Initialize an empty in-memory storage.

        The storage is implemented as a dictionary.
        """
        self.storage = {}

    def __getitem__(self, key: str) -> Any:
        """
        Get the item for the specified key.

        Parameters
        ----------
        key : str
            The key of the item to retrieve.

        Returns
        -------
        Any
            The value associated with the key.

        Raises
        ------
        KeyError
            If the key is not found in the storage.
        """
        return self.storage[key]

    def __setitem__(self, key: str, value: Any):
        """
        Set the item for the specified key.

        Parameters
        ----------
        key : str
            The key to associate with the value.
        value : Any
            The value to store.
        """
        self.storage[key] = value

    def __delitem__(self, key: str):
        """
        Delete the item for the specified key.

        Parameters
        ----------
        key : str
            The key of the item to delete.

        Raises
        ------
        KeyError
            If the key is not found in the storage.
        """
        del self.storage[key]

    def keys(self) -> Iterator[str]:
        """
        Yield the keys.

        Returns
        -------
        Iterator[str]
            An iterator yielding all keys in the storage.
        """
        return iter(self.storage.keys())

    def values(self) -> Iterator[Any]:
        """
        Yield the values.

        Returns
        -------
        Iterator[Any]
            An iterator yielding all values in the storage.
        """
        return iter(self.storage.values())

    def items(self) -> Iterator[tuple[str, Any]]:
        """
        Yield the pairs (key, value).

        Returns
        -------
        Iterator[tuple[str, Any]]
            An iterator yielding all (key, value) pairs in the storage.
        """
        return iter(self.storage.items())

    def __repr__(self) -> str:
        """
        Return a string representation of the storage.

        Returns
        -------
        str
            A string representation of the storage.
        """
        return "InMemoryStorage()"



"""CrossValidationItem class.

This class represents the output of a cross-validation workflow.
"""

from __future__ import annotations

import contextlib
import hashlib
from functools import cached_property
from typing import TYPE_CHECKING, Any

import numpy
import plotly.graph_objects
import plotly.io

from skore.item.item import Item, ItemTypeError

if TYPE_CHECKING:
    import sklearn.base


def plot_cross_validation(cv_results: dict) -> plotly.graph_objects.Figure:
    """Plot the result of a cross-validation run.

    Parameters
    ----------
    cv_results : dict
        The output of scikit-learn's cross_validate function.

    Returns
    -------
    plotly.graph_objects.Figure
        A plot of the cross-validation results
    """
    from datetime import timedelta

    import pandas
    import plotly.graph_objects as go

    _cv_results = cv_results.copy()

    with contextlib.suppress(KeyError):
        del _cv_results["indices"]
        del _cv_results["estimator"]

    df = pandas.DataFrame(_cv_results)

    # Move time columns to last and "test_score" to first
    if "fit_time" in df.columns:
        df.insert(len(df.columns) - 1, "fit_time", df.pop("fit_time"))
    if "score_time" in df.columns:
        df.insert(len(df.columns) - 1, "score_time", df.pop("score_time"))
    if "test_score" in df.columns:
        df.insert(0, "test_score", df.pop("test_score"))

    dict_labels = {
        "fit_time": "fit_time (seconds)",
        "score_time": "score_time (seconds)",
    }

    fig = go.Figure()

    for col_i, col_name in enumerate(df.columns):
        metric_name = dict_labels.get(col_name, col_name)
        bar_color = plotly.colors.qualitative.Plotly[
            col_i % len(plotly.colors.qualitative.Plotly)
        ]
        bar_x = [min(df.index) - 0.5, max(df.index) + 0.5]

        common_kwargs = dict(
            visible=True if col_i == 0 else "legendonly",
            legendgroup=f"group{col_i}",
            # If the metric is a duration (e.g. "fit_time"),
            # we show a different hover text
            hovertemplate=(
                "%{customdata}" f"<extra>{col_name} (timedelta)</extra>"
                if col_name.endswith("_time")
                else "%{y}"
            ),
            customdata=(
                [str(timedelta(seconds=x)) for x in df[col_name].values]
                if col_name.endswith("_time")
                else None
            ),
        )

        # Calculate statistics
        avg_value = df[col_name].mean()
        std_value = df[col_name].std()

        # Add all traces at once
        fig.add_traces(
            [
                # Bar trace
                go.Bar(
                    x=df.index,
                    y=df[col_name].values,
                    name=metric_name,
                    marker_color=bar_color,
                    showlegend=True,
                    **common_kwargs,
                ),
                # Mean line
                go.Scatter(
                    x=bar_x,
                    y=[avg_value, avg_value],
                    name=f"Average {metric_name}",
                    line=dict(dash="dash", color=bar_color),
                    showlegend=False,
                    mode="lines",
                    **common_kwargs,
                ),
                # +1 std line
                go.Scatter(
                    x=bar_x,
                    y=[avg_value + std_value, avg_value + std_value],
                    name=f"Average + 1 std. dev. {metric_name}",
                    line=dict(dash="dot", color=bar_color),
                    showlegend=False,
                    mode="lines",
                    **common_kwargs,
                ),
                # -1 std line
                go.Scatter(
                    x=bar_x,
                    y=[avg_value - std_value, avg_value - std_value],
                    name=f"Average - 1 std. dev. {metric_name}",
                    line=dict(dash="dot", color=bar_color),
                    showlegend=False,
                    mode="lines",
                    **common_kwargs,
                ),
            ]
        )

    fig.update_xaxes(tickmode="linear", dtick=1, title_text="Split number")
    fig.update_yaxes(title_text="Value")
    fig.update_layout(title_text="Cross-validation results for each split")

    return fig


def plot_cross_validation_aggregation(
    cv_results_items_versions: list[CrossValidationItem],
) -> plotly.graph_objects.Figure:
    """Plot the result of the aggregation of several cross-validation runs.

    Parameters
    ----------
    cv_results_items_versions : list[CrossValidationItem]
        A list of outputs of scikit-learn's cross_validate function.

    Returns
    -------
    plotly.graph_objects.Figure
        A plot of the aggregated cross-validation results
    """
    from datetime import timedelta

    import pandas
    import plotly.graph_objects as go

    _cv_results = cv_results_items_versions.copy()

    df = pandas.DataFrame([v.cv_results_serialized for v in _cv_results])
    df = df.apply(pandas.Series.explode)
    df = df.reset_index(names="run_number")

    # Move time columns to last and "test_score" to first
    if "fit_time" in df.columns:
        df.insert(len(df.columns) - 1, "fit_time", df.pop("fit_time"))
    if "score_time" in df.columns:
        df.insert(len(df.columns) - 1, "score_time", df.pop("score_time"))
    if "test_score" in df.columns:
        df.insert(1, "test_score", df.pop("test_score"))

    dict_labels = {
        "fit_time": "fit_time (seconds)",
        "score_time": "score_time (seconds)",
    }

    fig = go.Figure()

    for col_i, col_name in enumerate(df.columns[1:]):
        metric_name = dict_labels.get(col_name, col_name)
        bar_color = plotly.colors.qualitative.Plotly[
            col_i % len(plotly.colors.qualitative.Plotly)
        ]

        common_kwargs = dict(
            visible=True if col_i == 1 else "legendonly",
            legendgroup=f"group{col_i}",
            # If the metric is a duration (e.g. "fit_time"),
            # we show a different hover text
            hovertemplate=(
                "%{customdata}" f"<extra>{col_name} (timedelta)</extra>"
                if col_name.endswith("_time")
                else "%{y}"
            ),
            customdata=(
                [str(timedelta(seconds=x)) for x in df[col_name].values]
                if col_name.endswith("_time")
                else None
            ),
        )

        fig.add_traces(
            [
                go.Scatter(
                    x=df["run_number"],
                    y=df[col_name].values,
                    name=metric_name,
                    mode="markers",
                    marker_color=bar_color,
                    showlegend=True,
                    **common_kwargs,
                ),
            ]
        )

    fig.update_xaxes(tickmode="linear", dtick=1, title_text="Run number")
    fig.update_yaxes(title_text="Value")
    fig.update_layout(title_text="Cross-validation results for each run")

    return fig


def _hash_numpy(arr: numpy.ndarray) -> str:
    """Compute a hash string from a numpy array.

    Parameters
    ----------
    arr : numpy array
        The numpy array whose hash will be computed.

    Returns
    -------
    hash : str
        A hash corresponding to the input array.
    """
    return hashlib.sha256(bytes(memoryview(arr))).hexdigest()


# Data used for training, passed as input to scikit-learn
Data = Any
# Target used for training, passed as input to scikit-learn
Target = Any


class CrossValidationItem(Item):
    """
    A class to represent the output of a cross-validation workflow.

    This class encapsulates the output of scikit-learn's cross-validate function along
    with its creation and update timestamps.
    """

    def __init__(
        self,
        cv_results_serialized: dict,
        estimator_info: dict,
        X_info: dict,
        y_info: dict,
        plot_bytes: bytes,
        created_at: str | None = None,
        updated_at: str | None = None,
    ):
        """
        Initialize a CrossValidationItem.

        Parameters
        ----------
        cv_results_serialized : dict
            The dict output of scikit-learn's cross_validate function,
            in a form suitable for serialization.
        estimator_info : dict
            The estimator that was cross-validated.
        X_info : dict
            A summary of the data, input of scikit-learn's cross_validation function.
        y_info : dict
            A summary of the target, input of scikit-learn's cross_validation function.
        plot_bytes : bytes
            A plot of the cross-validation results, in the form of bytes.
        created_at : str
            The creation timestamp in ISO format.
        updated_at : str
            The last update timestamp in ISO format.
        """
        super().__init__(created_at, updated_at)

        self.cv_results_serialized = cv_results_serialized
        self.estimator_info = estimator_info
        self.X_info = X_info
        self.y_info = y_info
        self.plot_bytes = plot_bytes

    @classmethod
    def factory(
        cls,
        cv_results: dict,
        estimator: sklearn.base.BaseEstimator,
        X: Data,
        y: Target | None,
    ) -> CrossValidationItem:
        """
        Create a new CrossValidationItem instance.

        Parameters
        ----------
        cv_results : dict
            The dict output of scikit-learn's cross_validate function.
        estimator : sklearn.base.BaseEstimator,
            The estimator that was cross-validated.
        X
            The data, input of scikit-learn's cross_validation function.
        y
            The target, input of scikit-learn's cross_validation function.

        Returns
        -------
        CrossValidationItem
            A new CrossValidationItem instance.
        """
        if not isinstance(cv_results, dict):
            raise ItemTypeError(f"Type '{cv_results.__class__}' is not supported.")

        cv_results_serialized = {}
        for k, v in cv_results.items():
            if k == "estimator":
                continue
            if k == "indices":
                cv_results_serialized["indices"] = {
                    "train": tuple(arr.tolist() for arr in v["train"]),
                    "test": tuple(arr.tolist() for arr in v["test"]),
                }
            if isinstance(v, numpy.ndarray):
                cv_results_serialized[k] = v.tolist()

        estimator_info = {
            "name": estimator.__class__.__name__,
            "params": repr(estimator.get_params()),
        }

        y_array = y if isinstance(y, numpy.ndarray) else numpy.array(y)
        y_info = None if y is None else {"hash": _hash_numpy(y_array)}

        X_array = X if isinstance(X, numpy.ndarray) else numpy.array(X)
        X_info = {
            "nb_rows": X_array.shape[0],
            "nb_cols": X_array.shape[1],
            "hash": _hash_numpy(X_array),
        }

        # Keep plot itself as well as bytes so we can cache it
        plot = plot_cross_validation(cv_results_serialized)
        plot_bytes = plotly.io.to_json(plot, engine="json").encode("utf-8")

        instance = cls(
            cv_results_serialized=cv_results_serialized,
            estimator_info=estimator_info,
            X_info=X_info,
            y_info=y_info,
            plot_bytes=plot_bytes,
        )

        # Cache plot
        instance.plot = plot

        return instance

    @cached_property
    def plot(self):
        """A plot of the cross-validation results."""
        return plotly.io.from_json(self.plot_bytes.decode("utf-8"))


class CrossValidationAggregationItem(Item):
    """Aggregated outputs of several cross-validation workflow runs."""

    def __init__(
        self,
        plot_bytes: bytes,
        created_at: str | None = None,
        updated_at: str | None = None,
    ):
        """
        Initialize an CrossValidationAggregationItem.

        Parameters
        ----------
        plot_bytes : bytes
            A plot of the aggregated cross-validation results, in the form of bytes.
        created_at : str
            The creation timestamp in ISO format.
        updated_at : str
            The last update timestamp in ISO format.
        """
        super().__init__(created_at, updated_at)

        self.plot_bytes = plot_bytes

    @classmethod
    def factory(
        cls,
        cv_results_items_versions: list[CrossValidationItem],
    ) -> CrossValidationAggregationItem:
        """
        Create a new CrossValidationAggregationItem instance.

        Parameters
        ----------
        cv_results_items_versions: list[CrossValidationItem]
            A list of cross_validate items to be aggregated.

        Returns
        -------
        CrossValidationAggregationItem
            A new CrossValidationAggregationItem instance.
        """
        plot = plot_cross_validation_aggregation(cv_results_items_versions)

        plot_bytes = plotly.io.to_json(plot, engine="json").encode("utf-8")

        instance = cls(
            plot_bytes=plot_bytes,
        )

        # Cache plot
        instance.plot = plot

        return instance

    @cached_property
    def plot(self):
        """An aggregation plot of all the cross-validation results.

        Results are shown from the oldest to the current.
        """
        return plotly.io.from_json(self.plot_bytes.decode("utf-8"))



"""Configure logging and global settings."""

import logging

import rich.logging

from skore.project import Project, load
from skore.sklearn import cross_validate
from skore.utils._show_versions import show_versions

__all__ = [
    "Project",
    "cross_validate",
    "load",
    "show_versions",
]


class Handler(rich.logging.RichHandler):
    """A logging handler that renders output with Rich."""

    def get_level_text(self, record: rich.logging.LogRecord) -> rich.logging.Text:
        """Get the logger name and levelname from the record."""
        levelname = record.levelname
        name = record.name
        txt = f"<Logger {name} ({levelname})>"

        return rich.logging.Text.styled(
            txt.ljust(8),
            f"logging.level.{levelname.lower()}",
        )


formatter = logging.Formatter("%(message)s")
handler = Handler(markup=True)
handler.setFormatter(formatter)

logger = logging.getLogger(__name__)
logger.addHandler(handler)
logger.setLevel(logging.INFO)



"""Implement the "launch" command."""

import webbrowser
from contextlib import asynccontextmanager
from pathlib import Path
from typing import Union

import uvicorn
from fastapi import FastAPI

from skore.cli import logger
from skore.project import load
from skore.ui.app import create_app


class ProjectNotFound(Exception):
    """Project was not found."""

    project_path: Path


def __launch(project_name: Union[str, Path], port: int, open_browser: bool):
    """Launch the UI to visualize a project.

    Parameters
    ----------
    project_name : Path-like
        Name of the project to be created, or a relative or absolute path.
    port : int
        Port at which to bind the UI server.
    open_browser: bool
        Whether to automatically open a browser tab showing the UI.
    """
    project = load(project_name)

    @asynccontextmanager
    async def lifespan(app: FastAPI):
        if open_browser:
            webbrowser.open(f"http://localhost:{port}")
        yield

    app = create_app(project=project, lifespan=lifespan)

    try:
        # TODO: check port is free
        logger.info(
            f"Running skore UI from '{project_name}' at URL http://localhost:{port}"
        )
        uvicorn.run(app, port=port, log_level="error")
    except KeyboardInterrupt:
        logger.info("Closing skore UI")



"""Base class for all items in the project."""

from __future__ import annotations

import inspect
from abc import ABC, abstractmethod
from datetime import datetime, timezone
from functools import cached_property
from typing import Any, Optional


class ItemTypeError(Exception):
    """Item type exception.

    Exception raised when an attempt is made to convert an object to an Item, but the
    object's type is not supported.
    """


class Item(ABC):
    """
    Abstract base class for all items in the project.

    This class provides a common interface for all items, including
    creation and update timestamps.

    Parameters
    ----------
    created_at : str | None, optional
        The creation timestamp of the item. If None, the current time is used.
    updated_at : str | None, optional
        The last update timestamp of the item. If None, the current time is used.

    Attributes
    ----------
    created_at : str
        The creation timestamp of the item.
    updated_at : str
        The last update timestamp of the item.
    """

    def __init__(
        self,
        created_at: Optional[str] = None,
        updated_at: Optional[str] = None,
    ):
        now = datetime.now(tz=timezone.utc).isoformat()

        self.created_at = created_at or now
        self.updated_at = updated_at or now

    @classmethod
    @abstractmethod
    def factory(cls) -> Item:
        """
        Create and return a new instance of the Item.

        Returns
        -------
        Item
            A new instance of the Item.
        """

    @cached_property
    def __parameters__(self) -> dict[str, Any]:
        """
        Get the parameters of the Item instance.

        Returns
        -------
        dict[str, Any]
            A dictionary containing the parameters of the Item instance.
        """
        cls = self.__class__
        cls_parameters = inspect.signature(cls).parameters

        return {parameter: getattr(self, parameter) for parameter in cls_parameters}

    def __repr__(self) -> str:
        """Represent the item."""
        return f"{self.__class__.__name__}(...)"



"""Implement a repository for Views."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from skore.persistence.abstract_storage import AbstractStorage
    from skore.view.view import View


class ViewRepository:
    """
    A repository for managing storage and retrieval of Views.

    This class provides methods to get, put, and delete Views from a storage system.
    """

    def __init__(self, storage: AbstractStorage):
        """
        Initialize the ViewRepository with a storage system.

        Parameters
        ----------
        storage : AbstractStorage
            The storage system to be used by the repository.
        """
        self.storage = storage

    def get_view(self, key: str) -> View:
        """
        Retrieve the View from storage.

        Parameters
        ----------
        key : str
            A key at which to look for a View.

        Returns
        -------
        View
            The retrieved View.

        Raises
        ------
        KeyError
            When `key` is not present in the underlying storage.
        """
        return self.storage[key]

    def put_view(self, key: str, view: View):
        """
        Store a view in storage.

        Parameters
        ----------
        view : View
            The view to be stored.
        """
        self.storage[key] = view

    def delete_view(self, key: str):
        """Delete the view from storage."""
        del self.storage[key]

    def keys(self) -> list[str]:
        """
        Get all keys of items stored in the repository.

        Returns
        -------
        list[str]
            A list of all keys in the storage.
        """
        return list(self.storage.keys())



"""CLI for Skore."""

import argparse
import pathlib
from importlib.metadata import version

from skore.cli.create_project import __create
from skore.cli.launch_dashboard import __launch
from skore.cli.quickstart_command import __quickstart


def cli(args: list[str]):
    """CLI for Skore."""
    parser = argparse.ArgumentParser(prog="skore")

    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {version('skore')}"
    )

    subparsers = parser.add_subparsers(dest="subcommand")

    parser_launch = subparsers.add_parser("launch", help="Launch the web UI")
    parser_launch.add_argument(
        "project_name",
        help="the name or path of the project to open",
    )
    parser_launch.add_argument(
        "--port",
        type=int,
        help="the port at which to bind the UI server (default: %(default)s)",
        default=22140,
    )
    parser_launch.add_argument(
        "--open-browser",
        action=argparse.BooleanOptionalAction,
        help=(
            "whether to automatically open a browser tab showing the web UI "
            "(default: %(default)s)"
        ),
        default=True,
    )

    parser_create = subparsers.add_parser("create", help="Create a project")
    parser_create.add_argument(
        "project_name",
        nargs="?",
        help="the name or path of the project to create (default: %(default)s)",
        default="project",
    )
    parser_create.add_argument(
        "--working-dir",
        type=pathlib.Path,
        help=(
            "the directory relative to which the project name will be interpreted; "
            "default is the current working directory (mostly used for testing)"
        ),
        default=None,
    )

    subparsers.add_parser(
        "quickstart", help='Create a "project.skore" file and start the UI'
    )

    parsed_args: argparse.Namespace = parser.parse_args(args)

    if parsed_args.subcommand == "launch":
        __launch(
            project_name=parsed_args.project_name,
            port=parsed_args.port,
            open_browser=parsed_args.open_browser,
        )
    elif parsed_args.subcommand == "create":
        __create(
            project_name=parsed_args.project_name,
            working_dir=parsed_args.working_dir,
        )
    elif parsed_args.subcommand == "quickstart":
        __quickstart()
    else:
        parser.print_help()



"""Implement the "quickstart" command."""

from skore.cli import logger
from skore.cli.create_project import ProjectAlreadyExistsError, __create
from skore.cli.launch_dashboard import __launch


def __quickstart():
    """Quickstart a Skore project.

    Create it if it does not exist, then launch the web UI.

    Parameters
    ----------
    port : int
        Port at which to bind the UI server.
    """
    project_name = "project.skore"

    try:
        __create(project_name=project_name)
    except ProjectAlreadyExistsError:
        logger.info(
            f"Project file '{project_name}' already exists. Skipping creation step."
        )

    __launch(
        project_name="project.skore",
        port=22140,
        open_browser=True,
    )



"""Implement skore CLI."""

import logging

formatter = logging.Formatter("%(message)s")

console_handler = logging.StreamHandler()
console_handler.setLevel(logging.INFO)
console_handler.setFormatter(formatter)

logger = logging.getLogger(__name__)
logger.addHandler(console_handler)
logger.propagate = False



"""Implement the "create project" feature."""

import re
from pathlib import Path
from typing import Optional, Union

from skore.cli import logger
from skore.project import load
from skore.view.view import View


class InvalidProjectNameError(Exception):
    """The project name does not fit with one or more of the project name rules.

    - The project name must start with an alphanumeric character, and must not contain
    special characters other than '_' (underscore) and '-' (hyphen).
    - The project name must be at most 255 characters long (including ".skore").
    - The project name must not be a reserved OS file name.
    For example, CON, AUX, NUL... on Windows.
    """


def validate_project_name(project_name: str) -> tuple[bool, Optional[Exception]]:
    """Validate the project name (the part before ".skore").

    Returns `(True, None)` if validation succeeded and `(False, Exception(...))`
    otherwise.
    """
    # The project name (including the .skore extension) must be between 5 and 255
    # characters long.
    # FIXME: On Linux the OS already checks filename lengths
    if len(project_name) + len(".skore") > 255:
        return False, InvalidProjectNameError(
            "Project name length cannot exceed 255 characters."
        )

    # Reserved Names: The following names are reserved and cannot be used:
    # CON, PRN, AUX, NUL
    # COM1, COM2, COM3, COM4, COM5, COM6, COM7, COM8, COM9
    # LPT1, LPT2, LPT3, LPT4, LPT5, LPT6, LPT7, LPT8, LPT9
    reserved_patterns = "|".join(["CON", "PRN", "AUX", "NUL", r"COM\d+", r"LPT\d+"])
    if re.fullmatch(f"^({reserved_patterns})$", project_name):
        return False, InvalidProjectNameError(
            "Project name must not be a reserved OS filename."
        )

    # Allowed Characters:
    # Alphanumeric characters (a-z, A-Z, 0-9)
    # Underscore (_)
    # Hyphen (-)
    # Starting Character: The project name must start with an alphanumeric character.
    if not re.fullmatch(r"^[a-zA-Z0-9][a-zA-Z0-9_-]*$", project_name):
        return False, InvalidProjectNameError(
            "Project name must contain only alphanumeric characters, '_' and '-'."
        )

    # Case Sensitivity: File names are case-insensitive on Windows and case-sensitive
    # on Unix-based systems. The CLI should warn users about potential case conflicts
    # on Unix systems.

    return True, None


class ProjectCreationError(Exception):
    """Project creation failed."""


class ProjectAlreadyExistsError(Exception):
    """A project with this name already exists."""


class ProjectPermissionError(Exception):
    """Permissions in the directory do not allow creating a file."""


def __create(
    project_name: Union[str, Path], working_dir: Optional[Path] = None
) -> Path:
    """Create a project file named according to `project_name`.

    Parameters
    ----------
    project_name : Path-like
        Name of the project to be created, or a relative or absolute path.
    working_dir : Path or None
        If `project_name` is not an absolute path, it will be considered relative to
        `working_dir`. If `project_name` is an absolute path, `working_dir` will have
        no effect. If set to None (the default), `working_dir` will be re-set to the
        current working directory.

    Returns
    -------
    The project directory path
    """
    project_path = Path(project_name)

    # Remove trailing ".skore" if it exists to check the name is valid
    checked_project_name: str = project_path.name.split(".skore")[0]

    validation_passed, validation_error = validate_project_name(checked_project_name)
    if not validation_passed:
        raise ProjectCreationError(
            f"Unable to create project file '{project_path}'."
        ) from validation_error

    # The file must end with the ".skore" extension.
    # If not provided, it will be automatically appended.
    # If project name is an absolute path, we keep that path

    # NOTE: `working_dir` has no effect if `checked_project_name` is absolute
    if working_dir is None:
        working_dir = Path.cwd()
    project_directory = working_dir / (
        project_path.with_name(checked_project_name + ".skore")
    )

    try:
        project_directory.mkdir()
    except FileExistsError as e:
        raise ProjectAlreadyExistsError(
            f"Unable to create project file '{project_directory}' because a file "
            "with that name already exists. Please choose a different name or delete "
            "the existing file."
        ) from e
    except PermissionError as e:
        raise ProjectPermissionError(
            f"Unable to create project file '{project_directory}'. "
            "Please check your permissions for the current directory."
        ) from e
    except Exception as e:
        raise ProjectCreationError(
            f"Unable to create project file '{project_directory}'."
        ) from e

    # Once the main project directory has been created, created the nested directories

    items_dir = project_directory / "items"
    try:
        items_dir.mkdir()
    except Exception as e:
        raise ProjectCreationError(
            f"Unable to create project file '{items_dir}'."
        ) from e

    views_dir = project_directory / "views"
    try:
        views_dir.mkdir()
    except Exception as e:
        raise ProjectCreationError(
            f"Unable to create project file '{views_dir}'."
        ) from e

    p = load(project_directory)
    p.put_view("default", View(layout=[]))

    logger.info(f"Project file '{project_directory}' was successfully created.")
    return p



"""Entry point to Skore CLI."""

import sys

from skore.cli.cli import cli

if __name__ == "__main__":
    import rich.traceback

    # Display error tracebacks with Rich
    rich.traceback.install(show_locals=True)

    cli(sys.argv[1:])



"""ItemRepository for managing storage and retrieval of items.

This module provides the ItemRepository class, which is responsible for
storing, retrieving, and deleting items in a storage system.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from skore.item.item import Item
    from skore.persistence.abstract_storage import AbstractStorage


from skore.item.cross_validation_item import (
    CrossValidationAggregationItem,
    CrossValidationItem,
)
from skore.item.media_item import MediaItem
from skore.item.numpy_array_item import NumpyArrayItem
from skore.item.pandas_dataframe_item import PandasDataFrameItem
from skore.item.pandas_series_item import PandasSeriesItem
from skore.item.primitive_item import PrimitiveItem
from skore.item.sklearn_base_estimator_item import SklearnBaseEstimatorItem


class ItemRepository:
    """
    A repository for managing storage and retrieval of items.

    This class provides methods to get, put, and delete items from a storage system.

    Additionally, it keeps a record of all previously inserted items, by treating the
    storage as a map from keys to *lists of* values.
    """

    ITEM_CLASS_NAME_TO_ITEM_CLASS = {
        "MediaItem": MediaItem,
        "NumpyArrayItem": NumpyArrayItem,
        "PandasDataFrameItem": PandasDataFrameItem,
        "PandasSeriesItem": PandasSeriesItem,
        "PrimitiveItem": PrimitiveItem,
        "CrossValidationItem": CrossValidationItem,
        "CrossValidationAggregationItem": CrossValidationAggregationItem,
        "SklearnBaseEstimatorItem": SklearnBaseEstimatorItem,
    }

    def __init__(self, storage: AbstractStorage):
        """
        Initialize the ItemRepository with a storage system.

        Parameters
        ----------
        storage : AbstractStorage
            The storage system to be used by the repository.
        """
        self.storage = storage

    @staticmethod
    def __deconstruct_item(item: Item) -> dict:
        return {
            "item_class_name": item.__class__.__name__,
            "item": item.__parameters__,
        }

    @staticmethod
    def __construct_item(value) -> Item:
        item_class_name = value["item_class_name"]
        item_class = ItemRepository.ITEM_CLASS_NAME_TO_ITEM_CLASS[item_class_name]
        item = value["item"]

        return item_class(**item)

    def get_item(self, key) -> Item:
        """
        Get an item from storage.

        In practice, since each key is associated with a list of values,
        this will return the latest one.

        Parameters
        ----------
        key : Any
            The key used to identify the item in storage.

        Returns
        -------
        Item
            The retrieved item.
        """
        return ItemRepository.__construct_item(self.storage[key][-1])

    def get_item_versions(self, key) -> list[Item]:
        """
        Get all the versions of an item associated with `key` from the storage.

        The list is ordered from oldest to newest "put" date.

        Parameters
        ----------
        key : Any
            The key used to identify the item in storage.

        Returns
        -------
        list[Item]
            The retrieved list of items.
        """
        return [ItemRepository.__construct_item(value) for value in self.storage[key]]

    def put_item(self, key, item: Item) -> None:
        """
        Store an item in storage.

        This appends to a list of items previously associated with key `key`.

        Parameters
        ----------
        key : Any
            The key to use for storing the item.
        item : Item
            The item to be stored.
        """
        _item = ItemRepository.__deconstruct_item(item)

        if key in self.storage:
            items = self.storage[key]
            _item["item"]["created_at"] = items[0]["item"]["created_at"]

            self.storage[key] = items + [_item]
        else:
            self.storage[key] = [_item]

    def delete_item(self, key):
        """
        Delete an item from storage.

        Parameters
        ----------
        key : Any
            The key of the item to be deleted.
        """
        del self.storage[key]

    def keys(self) -> list[str]:
        """
        Get all keys of items stored in the repository.

        Returns
        -------
        list[str]
            A list of all keys in the storage.
        """
        return list(self.storage.keys())



"""Item types for the skore package."""

from __future__ import annotations

from contextlib import suppress
from typing import Any

from skore.item.cross_validation_item import CrossValidationItem
from skore.item.item import Item, ItemTypeError
from skore.item.item_repository import ItemRepository
from skore.item.media_item import MediaItem
from skore.item.numpy_array_item import NumpyArrayItem
from skore.item.pandas_dataframe_item import PandasDataFrameItem
from skore.item.pandas_series_item import PandasSeriesItem
from skore.item.primitive_item import PrimitiveItem
from skore.item.sklearn_base_estimator_item import SklearnBaseEstimatorItem


def object_to_item(object: Any) -> Item:
    """Transform an object into an Item."""
    for cls in (
        PrimitiveItem,
        PandasDataFrameItem,
        PandasSeriesItem,
        NumpyArrayItem,
        SklearnBaseEstimatorItem,
        MediaItem,
    ):
        with suppress(ImportError, ItemTypeError):
            # ImportError:
            #     The factories are responsible to import third-party libraries in a
            #     lazy way. If library is missing, an ImportError exception will
            #     automatically be thrown.
            # ItemTypeError:
            #     The factories are responsible for checking that parameters are of the
            #     correct type. If not, they throw a ItemTypeError exception.
            return cls.factory(object)

    raise NotImplementedError(f"Type '{object.__class__}' is not supported.")


__all__ = [
    "CrossValidationItem",
    "Item",
    "ItemRepository",
    "MediaItem",
    "NumpyArrayItem",
    "PandasDataFrameItem",
    "PandasSeriesItem",
    "PrimitiveItem",
    "SklearnBaseEstimatorItem",
    "object_to_item",
]



"""Project View models."""

from dataclasses import dataclass

# An ordered list of keys to display
Layout = list[str]


@dataclass
class View:
    """A view of a Project.

    Examples
    --------
    >>> View(layout=["a", "b"])
    View(...)
    """

    layout: Layout



"""MediaItem.

This module defines the MediaItem class, which represents media items.
"""

from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING, Any

from skore.item.item import Item, ItemTypeError

if TYPE_CHECKING:
    from altair.vegalite.v5.schema.core import TopLevelSpec as Altair
    from matplotlib.figure import Figure as Matplotlib
    from PIL.Image import Image as Pillow
    from plotly.basedatatypes import BaseFigure as Plotly


def lazy_is_instance(object: Any, cls_fullname: str) -> bool:
    """Return True if object is an instance of a class named `cls_fullname`."""
    return cls_fullname in {
        f"{cls.__module__}.{cls.__name__}" for cls in object.__class__.__mro__
    }


class MediaItem(Item):
    """
    A class to represent a media item.

    This class encapsulates various types of media along with metadata.
    """

    def __init__(
        self,
        media_bytes: bytes,
        media_encoding: str,
        media_type: str,
        created_at: str | None = None,
        updated_at: str | None = None,
    ):
        """
        Initialize a MediaItem.

        Parameters
        ----------
        media_bytes : bytes
            The raw bytes of the media content.
        media_encoding : str
            The encoding of the media content.
        media_type : str
            The MIME type of the media content.
        created_at : str, optional
            The creation timestamp in ISO format.
        updated_at : str, optional
            The last update timestamp in ISO format.
        """
        super().__init__(created_at, updated_at)

        self.media_bytes = media_bytes
        self.media_encoding = media_encoding
        self.media_type = media_type

    @classmethod
    def factory(cls, media, *args, **kwargs):
        """
        Create a new MediaItem instance.

        This is a generic factory method that dispatches to specific
        factory methods based on the type of media provided.

        Parameters
        ----------
        media : Any
            The media content to store.

        Raises
        ------
        NotImplementedError
            If the type of media is not supported.

        Returns
        -------
        MediaItem
            A new MediaItem instance.
        """
        if lazy_is_instance(media, "builtins.bytes"):
            return cls.factory_bytes(media, *args, **kwargs)
        if lazy_is_instance(media, "builtins.str"):
            return cls.factory_str(media, *args, **kwargs)
        if lazy_is_instance(media, "altair.vegalite.v5.schema.core.TopLevelSpec"):
            return cls.factory_altair(media, *args, **kwargs)
        if lazy_is_instance(media, "matplotlib.figure.Figure"):
            return cls.factory_matplotlib(media, *args, **kwargs)
        if lazy_is_instance(media, "PIL.Image.Image"):
            return cls.factory_pillow(media, *args, **kwargs)
        if lazy_is_instance(media, "plotly.basedatatypes.BaseFigure"):
            return cls.factory_plotly(media, *args, **kwargs)

        raise ItemTypeError(f"Type '{media.__class__}' is not supported.")

    @classmethod
    def factory_bytes(
        cls,
        media: bytes,
        media_encoding: str = "utf-8",
        media_type: str = "application/octet-stream",
    ) -> MediaItem:
        """
        Create a new MediaItem instance from bytes.

        Parameters
        ----------
        media : bytes
            The raw bytes of the media content.
        media_encoding : str, optional
            The encoding of the media content, by default "utf-8".
        media_type : str, optional
            The MIME type of the media content, by default "application/octet-stream".

        Returns
        -------
        MediaItem
            A new MediaItem instance.
        """
        return cls(
            media_bytes=media,
            media_encoding=media_encoding,
            media_type=media_type,
        )

    @classmethod
    def factory_str(cls, media: str, media_type: str = "text/markdown") -> MediaItem:
        """
        Create a new MediaItem instance from a string.

        Parameters
        ----------
        media : str
            The string content to store.
        media_type : str, optional
            The MIME type of the media content, by default "text/html".

        Returns
        -------
        MediaItem
            A new MediaItem instance.
        """
        media_bytes = media.encode("utf-8")

        return cls(
            media_bytes=media_bytes,
            media_encoding="utf-8",
            media_type=media_type,
        )

    @classmethod
    def factory_altair(cls, media: Altair) -> MediaItem:
        """
        Create a new MediaItem instance from an Altair chart.

        Parameters
        ----------
        media : Altair
            The Altair chart to store.

        Returns
        -------
        MediaItem
            A new MediaItem instance.
        """
        media_bytes = media.to_json().encode("utf-8")

        return cls(
            media_bytes=media_bytes,
            media_encoding="utf-8",
            media_type="application/vnd.vega.v5+json",
        )

    @classmethod
    def factory_matplotlib(cls, media: Matplotlib) -> MediaItem:
        """
        Create a new MediaItem instance from a Matplotlib figure.

        Parameters
        ----------
        media : Matplotlib
            The Matplotlib figure to store.

        Returns
        -------
        MediaItem
            A new MediaItem instance.
        """
        with BytesIO() as stream:
            media.savefig(stream, format="svg")
            media_bytes = stream.getvalue()

            return cls(
                media_bytes=media_bytes,
                media_encoding="utf-8",
                media_type="image/svg+xml",
            )

    @classmethod
    def factory_pillow(cls, media: Pillow) -> MediaItem:
        """
        Create a new MediaItem instance from a Pillow image.

        Parameters
        ----------
        media : Pillow
            The Pillow image to store.

        Returns
        -------
        MediaItem
            A new MediaItem instance.
        """
        with BytesIO() as stream:
            media.save(stream, format="png")
            media_bytes = stream.getvalue()

            return cls(
                media_bytes=media_bytes,
                media_encoding="utf-8",
                media_type="image/png",
            )

    @classmethod
    def factory_plotly(cls, media: Plotly) -> MediaItem:
        """
        Create a new MediaItem instance from a Plotly figure.

        Parameters
        ----------
        media : Plotly
            The Plotly figure to store.

        Returns
        -------
        MediaItem
            A new MediaItem instance.
        """
        import plotly.io

        media_bytes = plotly.io.to_json(media, engine="json").encode("utf-8")

        return cls(
            media_bytes=media_bytes,
            media_encoding="utf-8",
            media_type="application/vnd.plotly.v1+json",
        )



"""The definition of API routes to list project items and get them."""

import base64
from collections import defaultdict
from dataclasses import dataclass
from typing import Any

from fastapi import APIRouter, HTTPException, Request, status

from skore.item.cross_validation_item import (
    CrossValidationAggregationItem,
    CrossValidationItem,
)
from skore.item.media_item import MediaItem
from skore.item.numpy_array_item import NumpyArrayItem
from skore.item.pandas_dataframe_item import PandasDataFrameItem
from skore.item.pandas_series_item import PandasSeriesItem
from skore.item.primitive_item import PrimitiveItem
from skore.item.sklearn_base_estimator_item import SklearnBaseEstimatorItem
from skore.project import Project
from skore.view.view import Layout, View

router = APIRouter(prefix="/project")


@dataclass
class SerializedItem:
    """Serialized item."""

    media_type: str
    value: Any
    updated_at: str
    created_at: str


@dataclass
class SerializedProject:
    """Serialized project, to be sent to the skore-ui."""

    items: dict[str, list[SerializedItem]]
    views: dict[str, Layout]


def __serialize_project(project: Project) -> SerializedProject:
    items = defaultdict(list)

    for key in project.list_item_keys():
        for item in project.get_item_versions(key):
            if isinstance(item, PrimitiveItem):
                value = item.primitive
                media_type = "text/markdown"
            elif isinstance(item, NumpyArrayItem):
                value = item.array.tolist()
                media_type = "text/markdown"
            elif isinstance(item, PandasDataFrameItem):
                value = item.dataframe.to_dict(orient="tight")
                media_type = "application/vnd.dataframe+json"
            elif isinstance(item, PandasSeriesItem):
                value = item.series.to_list()
                media_type = "text/markdown"
            elif isinstance(item, SklearnBaseEstimatorItem):
                value = item.estimator_html_repr
                media_type = "application/vnd.sklearn.estimator+html"
            elif isinstance(item, MediaItem):
                value = base64.b64encode(item.media_bytes).decode()
                media_type = item.media_type
            elif isinstance(
                item, (CrossValidationItem, CrossValidationAggregationItem)
            ):
                value = base64.b64encode(item.plot_bytes).decode()
                media_type = "application/vnd.plotly.v1+json"
            else:
                raise ValueError(f"Item {item} is not a known item type.")

            items[key].append(
                SerializedItem(
                    media_type=media_type,
                    value=value,
                    updated_at=item.updated_at,
                    created_at=item.created_at,
                )
            )

    views = {key: project.get_view(key).layout for key in project.list_view_keys()}

    return SerializedProject(
        items=dict(items),
        views=views,
    )


@router.get("/items")
async def get_items(request: Request):
    """Serialize a project and send it."""
    project = request.app.state.project
    return __serialize_project(project)


@router.put("/views", status_code=status.HTTP_201_CREATED)
async def put_view(request: Request, key: str, layout: Layout):
    """Set the layout of the view corresponding to `key`.

    If the view corresponding to `key` does not exist, it will be created.
    """
    project: Project = request.app.state.project

    view = View(layout=layout)
    project.put_view(key, view)

    return __serialize_project(project)


@router.delete("/views", status_code=status.HTTP_202_ACCEPTED)
async def delete_view(request: Request, key: str):
    """Delete the view corresponding to `key`."""
    project: Project = request.app.state.project

    try:
        project.delete_view(key)
    except KeyError:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="View not found"
        ) from None

    return __serialize_project(project)



"""Define a Project."""

import logging
from pathlib import Path
from typing import Any, Optional, Union

from skore.item import (
    CrossValidationItem,
    Item,
    ItemRepository,
    MediaItem,
    NumpyArrayItem,
    PandasDataFrameItem,
    PandasSeriesItem,
    PrimitiveItem,
    SklearnBaseEstimatorItem,
    object_to_item,
)
from skore.persistence.disk_cache_storage import DirectoryDoesNotExist, DiskCacheStorage
from skore.view.view import View
from skore.view.view_repository import ViewRepository

logger = logging.getLogger(__name__)


class ProjectPutError(Exception):
    """One more key-value pairs could not be saved in the Project."""


class Project:
    """A project is a collection of items that are stored in a storage."""

    def __init__(
        self,
        item_repository: ItemRepository,
        view_repository: ViewRepository,
    ):
        self.item_repository = item_repository
        self.view_repository = view_repository

    def put(self, key: Union[str, dict[str, Any]], value: Optional[Any] = None):
        """Add one or more key-value pairs to the Project.

        If ``key`` is a string, then `put` adds the single ``key``-``value`` pair
        mapping to the Project.
        If ``key`` is a dict, it is interpreted as multiple key-value pairs to add to
        the Project.
        If an item with the same key already exists, its value is replaced by the new
        one.

        The dict format is the same as equivalent to running :func:`~skore.Project.put`
        for each individual key-value pair. In other words,

        .. code-block:: python

            project.put({"hello": 1, "goodbye": 2})

        is equivalent to

        .. code-block:: python

            project.put("hello", 1)
            project.put("goodbye", 2)

        In particular, this means that if some key-value pair is invalid
        (e.g. if a key is not a string, or a value's type is not supported),
        then all the key-value pairs up to the first offending key-value pair will
        be successfully inserted, *and then* an error will be raised.

        Parameters
        ----------
        key : str | dict[str, Any]
            The key to associate with ``value`` in the Project,
            or dict of key-value pairs to add to the Project.
        value : Any, optional
            The value to associate with ``key`` in the Project.
            If ``key`` is a dict, this argument is ignored.

        Raises
        ------
        ProjectPutError
            If the key-value pair(s) cannot be saved properly.
        """
        if isinstance(key, dict):
            for key_, value in key.items():
                self.put_one(key_, value)
        else:
            self.put_one(key, value)

    def put_one(self, key: str, value: Any):
        """Add a key-value pair to the Project.

        Parameters
        ----------
        key : str
            The key to associate with ``value`` in the Project. Must be a string.
        value : Any
            The value to associate with ``key`` in the Project.

        Raises
        ------
        ProjectPutError
            If the key-value pair cannot be saved properly.
        """
        try:
            item = object_to_item(value)
            self.put_item(key, item)
        except (NotImplementedError, TypeError) as e:
            raise ProjectPutError(
                "Key-value pair could not be inserted in the Project"
            ) from e

    def put_item(self, key: str, item: Item):
        """Add an Item to the Project."""
        if not isinstance(key, str):
            raise TypeError(
                f"Key must be a string; key '{key}' is of type '{type(key)}'"
            )

        self.item_repository.put_item(key, item)

    def get(self, key: str) -> Any:
        """Get the value corresponding to ``key`` from the Project.

        Parameters
        ----------
        key : str
            The key corresponding to the item to get.

        Raises
        ------
        KeyError
            If the key does not correspond to any item.
        """
        item = self.get_item(key)

        if isinstance(item, PrimitiveItem):
            return item.primitive
        elif isinstance(item, NumpyArrayItem):
            return item.array
        elif isinstance(item, PandasDataFrameItem):
            return item.dataframe
        elif isinstance(item, PandasSeriesItem):
            return item.series
        elif isinstance(item, SklearnBaseEstimatorItem):
            return item.estimator
        elif isinstance(item, CrossValidationItem):
            return item.cv_results_serialized
        elif isinstance(item, MediaItem):
            return item.media_bytes
        else:
            raise ValueError(f"Item {item} is not a known item type.")

    def get_item(self, key: str) -> Item:
        """Get the item corresponding to ``key`` from the Project.

        Parameters
        ----------
        key: str
            The key corresponding to the item to get.

        Returns
        -------
        item : Item
            The Item corresponding to key ``key``.

        Raises
        ------
        KeyError
            If the key does not correspond to any item.
        """
        return self.item_repository.get_item(key)

    def get_item_versions(self, key: str) -> list[Item]:
        """
        Get all the versions of an item associated with ``key`` from the Project.

        The list is ordered from oldest to newest "put" date.

        Parameters
        ----------
        key : str
            The key corresponding to the item to get.

        Returns
        -------
        list[Item]
            The list of items corresponding to key ``key``.

        Raises
        ------
        KeyError
            If the key does not correspond to any item.
        """
        return self.item_repository.get_item_versions(key)

    def list_item_keys(self) -> list[str]:
        """List all item keys in the Project.

        Returns
        -------
        list[str]
            The list of item keys. The list is empty if there is no item.
        """
        return self.item_repository.keys()

    def delete_item(self, key: str):
        """Delete the item corresponding to ``key`` from the Project.

        Parameters
        ----------
        key : str
            The key corresponding to the item to delete.

        Raises
        ------
        KeyError
            If the key does not correspond to any item.
        """
        self.item_repository.delete_item(key)

    def put_view(self, key: str, view: View):
        """Add a view to the Project."""
        self.view_repository.put_view(key, view)

    def get_view(self, key: str) -> View:
        """Get the view corresponding to ``key`` from the Project.

        Parameters
        ----------
        key : str
            The key of the item to get.

        Returns
        -------
        View
            The view corresponding to ``key``.

        Raises
        ------
        KeyError
            If the key does not correspond to any view.
        """
        return self.view_repository.get_view(key)

    def delete_view(self, key: str):
        """Delete the view corresponding to ``key`` from the Project.

        Parameters
        ----------
        key : str
            The key corresponding to the view to delete.

        Raises
        ------
        KeyError
            If the key does not correspond to any view.
        """
        return self.view_repository.delete_view(key)

    def list_view_keys(self) -> list[str]:
        """List all view keys in the Project.

        Returns
        -------
        list[str]
            The list of view keys. The list is empty if there is no view.
        """
        return self.view_repository.keys()


class ProjectLoadError(Exception):
    """Failed to load project."""


def load(project_name: Union[str, Path]) -> Project:
    """Load an existing Project given a project name or path."""
    # Transform a project name to a directory path:
    # - Resolve relative path to current working directory,
    # - Check that the file ends with the ".skore" extension,
    #    - If not provided, it will be automatically appended,
    # - If project name is an absolute path, we keep that path.

    path = Path(project_name).resolve()

    if path.suffix != ".skore":
        path = path.parent / (path.name + ".skore")

    if not Path(path).exists():
        raise ProjectLoadError(f"Project '{path}' does not exist: did you create it?")

    try:
        # FIXME should those hardcoded string be factorized somewhere ?
        item_storage = DiskCacheStorage(directory=Path(path) / "items")
        item_repository = ItemRepository(storage=item_storage)
        view_storage = DiskCacheStorage(directory=Path(path) / "views")
        view_repository = ViewRepository(storage=view_storage)
        project = Project(
            item_repository=item_repository,
            view_repository=view_repository,
        )
    except DirectoryDoesNotExist as e:
        missing_directory = e.args[0].split()[1]
        raise ProjectLoadError(
            f"Project '{path}' is corrupted: "
            f"directory '{missing_directory}' should exist. "
            "Consider re-creating the project."
        ) from e

    return project



from pathlib import Path
from contextlib import suppress

from hatchling.metadata.plugin.interface import MetadataHookInterface


class MetadataHook(MetadataHookInterface):
    def update(self, metadata):
        license = Path(self.root, self.config["license-file"]).read_text()
        readme = Path(self.root, self.config["readme-file"]).read_text()
        version = self.config["version-default"]

        with suppress(FileNotFoundError):
            version = Path(self.root, "VERSION.txt").read_text()

        metadata["license"] = {"text": license, "content-type": "text/plain"}
        metadata["readme"] = {"text": readme, "content-type": "text/markdown"}
        metadata["version"] = version



"""cross_validate function.

This function implements a wrapper over scikit-learn's
`cross_validate <https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.cross_validate.html#sklearn.model_selection.cross_validate>`_
function in order to enrich it with more information and enable more analysis.
"""

import contextlib
from typing import Literal, Optional

from skore.item.cross_validation_item import (
    CrossValidationAggregationItem,
    CrossValidationItem,
)
from skore.project import Project


def _find_ml_task(
    estimator, y
) -> Literal[
    "binary-classification",
    "multiclass-classification",
    "regression",
    "clustering",
    "unknown",
]:
    """Guess the ML task being addressed based on an estimator and a target array.

    Parameters
    ----------
    estimator : sklearn.base.BaseEstimator
        An estimator.
    y : numpy.ndarray
        A target vector.

    Returns
    -------
    Literal["classification", "regression", "clustering", "unknown"]
        The guess of the kind of ML task being performed.
    """
    import sklearn.utils.multiclass
    from sklearn.base import is_classifier, is_regressor

    if y is None:
        # NOTE: The task might not be clustering
        return "clustering"

    if is_regressor(estimator):
        return "regression"

    type_of_target = sklearn.utils.multiclass.type_of_target(y)

    if is_classifier(estimator):
        if type_of_target == "binary":
            return "binary-classification"

        if type_of_target == "multiclass":
            return "multiclass-classification"

    if type_of_target == "unknown":
        return "unknown"

    if "continuous" in type_of_target:
        return "regression"

    return "classification"


def _get_scorers_to_add(estimator, y) -> list[str]:
    """Get a list of scorers based on `estimator` and `y`.

    Parameters
    ----------
    estimator : sklearn.base.BaseEstimator
        An estimator.
    y : numpy.ndarray
        A target vector.

    Returns
    -------
    scorers_to_add : list[str]
        A list of scorers
    """
    ml_task = _find_ml_task(estimator, y)

    # Add scorers based on the ML task
    if ml_task == "regression":
        return ["r2", "neg_root_mean_squared_error"]
    if ml_task == "binary-classification":
        return ["roc_auc", "neg_brier_score", "recall", "precision"]
    if ml_task == "multiclass-classification":
        if hasattr(estimator, "predict_proba"):
            return [
                "recall_weighted",
                "precision_weighted",
                "roc_auc_ovr_weighted",
                "neg_log_loss",
            ]
        return ["recall_weighted", "precision_weighted"]
    return []


def _add_scorers(scorers, scorers_to_add):
    """Expand `scorers` with more scorers.

    The type of the resulting scorers object is dependent on the type of the input
    scorers:
    - If `scorers` is a dict, then extra scorers are added to the dict;
    - If `scorers` is a string or None, then it is converted to a dict and extra scorers
    are added to the dict;
    - If `scorers` is a list or tuple, then it is converted to a dict and extra scorers
    are added to the dict;
    - If `scorers` is a callable, then a new callable is created that
    returns a dict with the user-defined score as well as the scorers to add.
    In case the user-defined dict contains a metric with a name conflicting with the
    metrics we add, the user-defined metric always wins.

    Parameters
    ----------
    scorers : any type that is accepted by scikit-learn's cross_validate
        The scorer(s) to expand.
    scorers_to_add : list[str]
        The scorers to be added.

    Returns
    -------
    new_scorers : dict or callable
        The scorers after adding `scorers_to_add`.
    added_scorers : Iterable[str]
        The scorers that were actually added (i.e. the ones that were not already
        in `scorers`).
    """
    if scorers is None or isinstance(scorers, str):
        new_scorers, added_scorers = _add_scorers({"score": scorers}, scorers_to_add)
    elif isinstance(scorers, (list, tuple)):
        new_scorers, added_scorers = _add_scorers(
            {s: s for s in scorers}, scorers_to_add
        )
    elif isinstance(scorers, dict):
        new_scorers = {s: s for s in scorers_to_add} | scorers
        added_scorers = set(scorers_to_add) - set(scorers)
    elif callable(scorers):
        from sklearn.metrics import check_scoring
        from sklearn.metrics._scorer import _MultimetricScorer

        internal_scorer = _MultimetricScorer(
            scorers={
                s: check_scoring(estimator=None, scoring=s) for s in scorers_to_add
            }
        )

        def new_scorer(estimator, X, y) -> dict:
            scores = scorers(estimator, X, y)
            if isinstance(scores, dict):
                return internal_scorer(estimator, X, y) | scores
            return internal_scorer(estimator, X, y) | {"score": scores}

        new_scorers = new_scorer

        # In this specific case, we can't know if there is overlap between the
        # user-defined scores and ours, so we take the least risky option
        # which is to say we added nothing; that way, we won't remove anything
        # after cross-validation is computed
        added_scorers = []

    return new_scorers, added_scorers


def _strip_cv_results_scores(cv_results: dict, added_scorers: list[str]) -> dict:
    """Remove information about `added_scorers` in `cv_results`.

    Parameters
    ----------
    cv_results : dict
        A dict of the form returned by scikit-learn's cross_validate function.
    added_scorers : list[str]
        A list of scorers in `cv_results` which should be removed.

    Returns
    -------
    dict
        A new cv_results dict, with the specified scorers information removed.
    """
    # Takes care both of train and test scores
    return {
        k: v
        for k, v in cv_results.items()
        if not any(added_scorer in k for added_scorer in added_scorers)
    }


def cross_validate(*args, project: Optional[Project] = None, **kwargs) -> dict:
    """Evaluate estimator by cross-validation and output UI-friendly object.

    This function wraps scikit-learn's :func:`~sklearn.model_selection.cross_validate`
    function, to provide more context and facilitate the analysis.
    As such, the arguments are the same as scikit-learn's ``cross_validate`` function.

    The dict returned by this function is a strict super-set of the one returned by
    scikit-learn's :func:`~sklearn.model_selection.cross_validate`.

    For a user guide and in-depth example, see :ref:`example_cross_validate`.

    Parameters
    ----------
    *args
        Positional arguments accepted by scikit-learn's
        :func:`~sklearn.model_selection.cross_validate`,
        such as ``estimator`` and ``X``.
    project : Project, optional
        A project to save cross-validation data into. If None, no save is performed.
    **kwargs
        Additional keyword arguments accepted by scikit-learn's
        :func:`~sklearn.model_selection.cross_validate`.

    Returns
    -------
    cv_results : dict
        A dict of the form returned by scikit-learn's
        :func:`~sklearn.model_selection.cross_validate` function.

    Examples
    --------
    >>> def prepare_cv():
    ...     from sklearn import datasets, linear_model
    ...     diabetes = datasets.load_diabetes()
    ...     X = diabetes.data[:150]
    ...     y = diabetes.target[:150]
    ...     lasso = linear_model.Lasso()
    ...     return lasso, X, y

    >>> project = skore.load("project.skore")  # doctest: +SKIP
    >>> lasso, X, y = prepare_cv()  # doctest: +SKIP
    >>> cross_validate(lasso, X, y, cv=3, project=project)  # doctest: +SKIP
    {'fit_time': array(...), 'score_time': array(...), 'test_score': array(...)}
    """
    import sklearn.model_selection

    # Recover specific arguments
    estimator = args[0] if len(args) >= 1 else kwargs.get("estimator")
    X = args[1] if len(args) >= 2 else kwargs.get("X")
    y = args[2] if len(args) == 3 else kwargs.get("y")

    try:
        scorers = kwargs.pop("scoring")
    except KeyError:
        scorers = None

    # Extend scorers with other relevant scorers
    scorers_to_add = _get_scorers_to_add(estimator, y)
    new_scorers, added_scorers = _add_scorers(scorers, scorers_to_add)

    cv_results = sklearn.model_selection.cross_validate(
        *args, **kwargs, scoring=new_scorers
    )

    cross_validation_item = CrossValidationItem.factory(cv_results, estimator, X, y)

    if project is not None:
        try:
            cv_results_history = project.get_item_versions("cross_validation")
        except KeyError:
            cv_results_history = []

        agg_cross_validation_item = CrossValidationAggregationItem.factory(
            cv_results_history + [cross_validation_item]
        )

        project.put_item("cross_validation_aggregated", agg_cross_validation_item)
        project.put_item("cross_validation", cross_validation_item)

    # If in a IPython context (e.g. Jupyter notebook), display the plot
    with contextlib.suppress(ImportError):
        from IPython.core.interactiveshell import InteractiveShell
        from IPython.display import display

        if InteractiveShell.initialized():
            display(cross_validation_item.plot)

    # Remove information related to our scorers, so that our return value is
    # the same as sklearn's
    stripped_cv_results = _strip_cv_results_scores(cv_results, added_scorers)

    # Add explicit metric to result (rather than just "test_score")
    if isinstance(scorers, str):
        if kwargs.get("return_train_score") is not None:
            stripped_cv_results[f"train_{scorers}"] = stripped_cv_results["train_score"]
        stripped_cv_results[f"test_{scorers}"] = stripped_cv_results["test_score"]

    return stripped_cv_results



"""Abstract storage interface."""

from abc import ABC, abstractmethod
from collections.abc import Iterator
from typing import Any


class AbstractStorage(ABC):
    """Persist data in a storage."""

    @abstractmethod
    def __getitem__(self, key: str) -> Any:
        """
        Get the item for the specified key.

        Parameters
        ----------
        key : str
            The key of the item to retrieve.

        Returns
        -------
        Any
            The value associated with the key.

        Raises
        ------
        KeyError
            If the key is not found in the storage.
        """

    @abstractmethod
    def __setitem__(self, key: str, value: Any):
        """
        Set the item for the specified key.

        Parameters
        ----------
        key : str
            The key to associate with the value.
        value : Any
            The value to store.
        """

    @abstractmethod
    def __delitem__(self, key: str):
        """
        Delete the item for the specified key.

        Parameters
        ----------
        key : str
            The key of the item to delete.

        Raises
        ------
        KeyError
            If the key is not found in the storage.
        """

    @abstractmethod
    def keys(self) -> Iterator[str]:
        """
        Yield the keys in the storage.

        Returns
        -------
        Iterator[str]
            An iterator yielding all keys in the storage.
        """

    @abstractmethod
    def values(self) -> Iterator[Any]:
        """
        Yield the values in the storage.

        Returns
        -------
        Iterator[Any]
            An iterator yielding all values in the storage.
        """

    @abstractmethod
    def items(self) -> Iterator[tuple[str, Any]]:
        """
        Yield the pairs (key, value) in the storage.

        Returns
        -------
        Iterator[tuple[str, Any]]
            An iterator yielding all (key, value) pairs in the storage.
        """

    def __contains__(self, key: str) -> bool:
        """
        Return True if the storage has the specified key, else False.

        Parameters
        ----------
        key : str
            The key to check for existence in the storage.

        Returns
        -------
        bool
            True if the key is in the storage, else False.
        """
        return key in self.keys()

    def __len__(self) -> int:
        """
        Return the number of items in the storage.

        Returns
        -------
        int
            The number of items in the storage.
        """
        return len(list(self.keys()))

    def __iter__(self) -> Iterator[str]:
        """
        Yield the keys in the storage.

        Returns
        -------
        Iterator[str]
            An iterator yielding all keys in the storage.
        """
        return self.keys()
